#include "autarky.h"
#include "backtrack.h"
#include "decide.h"
#include "internal.h"
#include "logging.h"
#include "print.h"
#include "rephase.h"
#include "report.h"
#include "terminate.h"
#include "walk.h"

#include <inttypes.h>

void
kissat_reset_rephased (kissat * solver)
{
  LOG ("reset rephase type and counter");
  solver->rephased.type = 0;
}

void
kissat_reset_best_assigned (kissat * solver)
{
  if (!solver->best_assigned)
    return;
  LOG ("resetting best assigned %u", solver->best_assigned);
  solver->best_assigned = 0;
}

void
kissat_reset_target_assigned (kissat * solver)
{
  if (!solver->target_assigned)
    return;
  LOG ("resetting target assigned %u", solver->target_assigned);
  solver->target_assigned = 0;
}

bool
kissat_rephasing (kissat * solver)
{
  if (!solver->enabled.rephase)
    return false;
  return CONFLICTS > solver->limits.rephase.conflicts;
}

static char
rephase_best (kissat * solver)
{
  assert (GET_OPTION (rephasebest) || GET_OPTION (rephasewalking));
  for (all_phases (p))
    if (p->best)
      p->saved = p->best;
  INC (rephased_best);
  return 'B';
}

char
kissat_rephase_best (kissat * solver)
{
  return rephase_best (solver);
}

static char
rephase_flipped (kissat * solver)
{
  assert (GET_OPTION (rephaseflipped));
  for (all_phases (p))
    p->saved *= -1;
  INC (rephased_flipped);
  return 'F';
}

static char
rephase_original (kissat * solver)
{
  assert (GET_OPTION (rephaseoriginal));
  const value value = INITIAL_PHASE;
  for (all_phases (p))
    p->saved = value;
  INC (rephased_original);
  return 'O';
}

static char
rephase_inverted (kissat * solver)
{
  assert (GET_OPTION (rephaseinverted));
  const value value = -INITIAL_PHASE;
  for (all_phases (p))
    p->saved = value;
  INC (rephased_inverted);
  return 'I';
}

static char
rephase_random (kissat * solver)
{
  assert (GET_OPTION (rephaserandom));
  for (all_phases (p))
    p->saved = (kissat_pick_bool (&solver->random) ? 1 : -1);
  INC (rephased_random);
  return '#';
}

static char
rephase_walking (kissat * solver)
{
  assert (GET_OPTION (rephasewalking));
  STOP (rephase);
  char res = kissat_walk (solver);
  if (res == 'W')
    kissat_autarky (solver);
  START (rephase);
  if (!res)
    res = rephase_best (solver);
  if (res == 'W')
    INC (rephased_walking);
  else
    {
      assert (res == 'B');
      INC (rephased_best);
    }
  return res;
}

static void
reset_phases (kissat * solver)
{
  kissat_clear_target_phases (solver);
  kissat_reset_target_assigned (solver);
  const uint64_t count = solver->rephased.count++;

  const bool inverted = GET_OPTION (rephaseinverted);
  const bool original = GET_OPTION (rephaseoriginal);

  char type = 0;

  if (!count && original)
    type = rephase_original (solver);
  else if (!count && inverted)
    type = rephase_inverted (solver);
  else if (count == 1 && original && inverted)
    type = rephase_inverted (solver);
  else
    {
      const bool best = GET_OPTION (rephasebest);
      const bool flipped = GET_OPTION (rephaseflipped);
      const bool random = GET_OPTION (rephaserandom);
      const bool walking = GET_OPTION (rephasewalking);

      char (*functions[12])(kissat *);
      uint64_t candidates = 0;

#define PUSH1(NAME) \
      if (NAME) \
	{ \
	  assert (candidates < sizeof functions / sizeof *functions); \
	  functions[candidates++] = rephase_ ## NAME; \
	}

#define PUSH3(NAME) \
      if (NAME) \
	{ \
	  PUSH1 (best); \
	  PUSH1 (walking); \
	  PUSH1 (NAME); \
	}

      PUSH3 (original);
      PUSH3 (inverted);
      PUSH3 (random);
      PUSH3 (flipped);

      if (candidates)
	{
	  const uint64_t select = (count - original - inverted) % candidates;
	  type = functions [select] (solver);
	}
      else
	{
	  assert (best);
	  type = rephase_best (solver);
	}
    }
#ifndef QUIET
  const char *type_as_string = 0;
  switch (type)
    {
#define REPHASE(NAME, TYPE) \
      case TYPE: \
        type_as_string = #NAME; \
	break;
      REPHASES
#undef REPHASE
    }
  assert (type_as_string);
  kissat_phase (solver, "rephase", GET (rephased),
		"%s phases in %s search mode",
		type_as_string, solver->stable ? "stable" : "focused");
#endif
  LOG ("remember last rephase type '%c'", type);
  solver->rephased.type = type;
  solver->rephased.last = CONFLICTS;

  UPDATE_CONFLICT_LIMIT (rephase, rephased, LINEAR, false);
}

void
kissat_rephase (kissat * solver)
{
  kissat_backtrack_propagate_and_flush_trail (solver);
  assert (!solver->inconsistent);
  kissat_autarky (solver);
  if (TERMINATED (10))
    return;
  START (rephase);
  INC (rephased);
  REPORT (1, '~');
  reset_phases (solver);
  STOP (rephase);
}
